% classify(ff_features,root_reports,set_sizes,features,C);
%
% Generalization accuracy by leave-one-out (after feature extraction)
%   C: structure as follows:
% Majority vote:           C.classifier = 'majvote'
% Gaussian Naive Bayes:    C.classifier = 'gnb'
% K-Nearest Neighbors:     C.classifier = 'knn', C.neighbors = integer array
%                          Assumes euclidean distance
% Fisher linear discrim.:  C.classifier = 'fld', C.ridge = double
% Sparse logistic regres.: C.classifier = 'slogreg', C.rho = double array
% Linear Support Vec.Mac.: C.classifier = 'svmlin'
% RBF Support Vec.Machine: C.classifier = 'svmrbf', C.gamma = double array
% Adaboost:                C.classifier = 'adaboost', C.iterations = integer array
%
% Author: Jean Honorio
% E-mail: jhonorio@csail.mit.edu
function classify(ff_features,root_reports,tasks,set_sizes,features,C);

% Set classifier
warning off;
if strcmp(C.classifier,'majvote')
    classifier = @(test_data,train_data,train_labels,fold_subject) ...
        majvote(test_data,train_data,train_labels);
    param = [];
elseif strcmp(C.classifier,'majvote1')
    classifier = @(test_data,train_data,train_labels,fold_subject) ...
        majvote(test_data,train_data,train_labels,1);
    param = [];
elseif strcmp(C.classifier,'gnb')
    classifier = @(test_data,train_data,train_labels,fold_subject) ...
        gnb(test_data,train_data,train_labels);
    param = [];
elseif strcmp(C.classifier,'fld')
    classifier = @(test_data,train_data,train_labels,fold_subject) ...
        ridge_fld(test_data,train_data,train_labels,C.ridge);
    param = C.ridge;
elseif strcmp(C.classifier,'knn')
    classifier = @(test_data,train_data,train_labels,fold_subject) ...
        ncv_knn(test_data,train_data,train_labels,fold_subject,C.neighbors);
    param = C.neighbors;
elseif strcmp(C.classifier,'slogreg')
    classifier = @(test_data,train_data,train_labels,fold_subject) ...
        ncv_slogreg(test_data,train_data,train_labels,fold_subject,C.rho);
    param = C.rho;
elseif strcmp(C.classifier,'svmlin')
    K.kernel = 'linear';
    classifier = @(test_data,train_data,train_labels,fold_subject) ...
        svm(test_data,train_data,train_labels,K);
    param = [];
elseif strcmp(C.classifier,'svmrbf')
    classifier = @(test_data,train_data,train_labels,fold_subject) ...
        ncv_svmrbf(test_data,train_data,train_labels,fold_subject,C.gamma);
    param = C.gamma;
elseif strcmp(C.classifier,'adaboost')
    classifier = @(test_data,train_data,train_labels,fold_subject) ...
        ncv_adaboost(test_data,train_data,train_labels,fold_subject,C.iterations);
    param = C.iterations;
else
    error('Not a valid classifier');
end

% Create output directory if it doesn't exist
mkdir(root_reports);
if isempty(param)
    classifier_name = C.classifier;
else
    classifier_name = [C.classifier,sprintf('_%g',param)];
end
report_file = fullfile(root_reports,['features_',features,'_classifier_',classifier_name,'.txt']);

% Load extracted features
file_name = ff_features;
file_name = strrep(file_name,'[features]',features);
fprintf('Loading %s...\n',file_name);
extract = load(file_name);

% Check fold array
if ~isfield(extract,'fold_subject')
    extract.fold_subject = 1:length(extract.subjects);
end
check_fold_array(extract.fold_subject);

% Extract only required tasks
idx_tasks = [];
for t = 1:length(tasks)
    idx_t = find(strcmp(tasks{t},extract.tasks));
    if isempty(idx_t)
        error('Task %s not found',tasks{t});
    end
    idx_tasks = [idx_tasks, idx_t];
end
extract.folds = extract.folds(:,idx_tasks);
extract.tasks = extract.tasks(idx_tasks);

% For each possible size set and set of features
for set_size = set_sizes
    sets = nchoosek(1:length(extract.tasks),set_size);
    for set = 1:size(sets,1)

        % Classify
		succeeds = 0;
		fails = 0;
		failed_subjects = {};
        params_used = [];
        for s = 1:max(extract.fold_subject)
            fprintf('Fold %d\n',s);
            idx_sf = find(extract.fold_subject == s);
            fprintf('Subject');
            for sf = idx_sf
                fprintf(' %s',extract.subjects{sf});
            end
            fprintf('\n');
            idx_s = setdiff(1:length(extract.subjects),idx_sf);
            
            % Collect train and testing sets
            X_train = [];
            X_test = [];
            for t = sets(set,:)
                if size(extract.folds,1) == 1
                    fold = 1;
                else
                    fold = s;
                end
                X_train = [X_train extract.folds(fold,t).features(idx_s,:)];
                X_test = [X_test extract.folds(fold,t).features(idx_sf,:)];
            end
            label_train = extract.labels(idx_s);
            fold_train = extract.fold_subject(idx_s);
            idx_gt = find(fold_train > s);
            fold_train(idx_gt) = fold_train(idx_gt) - 1;
			check_fold_array(fold_train); %%% Might be unnecesary for some methods
            label_test = extract.labels(idx_sf);
            label_pred = classifier(X_test,X_train,label_train,fold_train);
			for s2 = 1:length(idx_sf)
				if label_pred(s2) == label_test(s2)
					succeeds = succeeds + 1;
				else
					fails = fails + 1;
					failed_subjects{length(failed_subjects)+1} = extract.subjects{idx_sf(s2)};
				end
			end
            if length(label_pred) > length(idx_sf)
                params_used = [params_used label_pred(length(idx_sf)+1:end)];
            end
        end

        % Add results to report
        fp = fopen(report_file,'a');
        if fp < 0
            error('Can''t open %s',report_file);
        end
        fprintf(fp,'Conditions:');
        fprintf('Conditions:');
        for t = sets(set,:)
            fprintf(fp,' %s',extract.tasks{t});
            fprintf(' %s',extract.tasks{t});
        end
        fprintf(fp,'\n');
        fprintf('\n');
		accuracy = 100*succeeds/(succeeds+fails);
		fprintf(fp,'Accuracy: %0.1f%%\n',accuracy);
		fprintf('Accuracy: %0.1f%%\n',accuracy);
		fprintf(fp,'Failed subjects: ');
		fprintf('Failed subjects: ');
		for i = 1:length(failed_subjects)
			fprintf(fp,'%s ',failed_subjects{i});
			fprintf('%s ',failed_subjects{i});
		end
		fprintf(fp,'\n');
		fprintf('\n');
        if ~isempty(params_used)
            fprintf(fp,'Parameters used: ');
            fprintf('Parameters used: ');
            fprintf(fp,'%d ',params_used);
            fprintf('%d ',params_used);
            fprintf(fp,'\n');
            fprintf('\n');
        end
        fprintf(fp,'\n');
        fprintf('\n');
        fclose(fp);
    end
end
fprintf('Done!\n');

% KNN classifier, pick k by nested cross validation
function test_labels = ncv_knn(test_data,train_data,train_labels,fold_subject,neighbors);

if length(neighbors) > 1
    good_pred = zeros(max(fold_subject),length(neighbors));
    for s = 1:max(fold_subject)
        idx_sf = find(fold_subject == s);
        idx_s = setdiff(1:length(train_labels),idx_sf);
        for n = 1:length(neighbors)
            if neighbors(n) <= length(idx_s)
                good_pred(s,n) = sum(train_labels(idx_sf).' == knnclassify(train_data(idx_sf,:),train_data(idx_s,:),train_labels(idx_s),neighbors(n)));
            end
        end
    end
    [tmp n] = max(sum(good_pred,1));
else
    n = 1;
end
test_labels = knnclassify(test_data,train_data,train_labels,neighbors(n));
test_labels = [reshape(test_labels,1,[]), neighbors(n)];

% sparse logistic regression, pick rho by nested cross validation
function test_labels = ncv_slogreg(test_data,train_data,train_labels,fold_subject,rho);

if length(rho) > 1
    good_pred = zeros(max(fold_subject),length(rho));
    for s = 1:max(fold_subject)
        idx_sf = find(fold_subject == s);
        idx_s = setdiff(1:length(train_labels),idx_sf);
        for r = 1:length(rho)
            good_pred(s,r) = sum(train_labels(idx_sf) == slogreg(train_data(idx_sf,:),train_data(idx_s,:),train_labels(idx_s),rho(r)));
        end
    end
    [tmp r] = max(sum(good_pred,1));
else
    r = 1;
end
test_labels = slogreg(test_data,train_data,train_labels,rho(r));
test_labels = [reshape(test_labels,1,[]), rho(r)];

% SVM/RBF classifier, pick gamma by nested cross validation
function test_labels = ncv_svmrbf(test_data,train_data,train_labels,fold_subject,gamma);

K.kernel = 'rbf';
if length(gamma) > 1
    good_pred = zeros(max(fold_subject),length(gamma));
    for s = 1:max(fold_subject)
        idx_sf = find(fold_subject == s);
        idx_s = setdiff(1:length(train_labels),idx_sf);
        for g = 1:length(gamma)
            K.gamma = gamma(g);
            good_pred(s,g) = sum(train_labels(idx_sf).' == svm(train_data(idx_sf,:),train_data(idx_s,:),train_labels(idx_s),K));
        end
    end
    [tmp g] = max(sum(good_pred,1));
else
    g = 1;
end
K.gamma = gamma(g);
test_labels = svm(test_data,train_data,train_labels,K);
test_labels = [reshape(test_labels,1,[]), K.gamma];

% Adaboost classifier, pick iterations by nested cross validation
function test_labels = ncv_adaboost(test_data,train_data,train_labels,fold_subject,iterations);

if length(iterations) > 1
    good_pred = zeros(max(fold_subject),length(iterations));
    for s = 1:max(fold_subject)
        idx_sf = find(fold_subject == s);
        idx_s = setdiff(1:length(train_labels),idx_sf);
        good_pred(s,:) = sum(train_labels(idx_sf).'*ones(1,length(iterations)) == adaboost(train_data(idx_sf,:),train_data(idx_s,:),train_labels(idx_s),iterations));
    end
    [tmp i] = max(sum(good_pred,1));
else
    i = 1;
end
test_labels = adaboost(test_data,train_data,train_labels,iterations(i));
test_labels = [reshape(test_labels,1,[]), iterations(i)];
