% extract_features(ff_contrasts,root_output,tasks,contrasts,...
%                  subjects,labels,fold_subject,scaling,C);
%
% Extract features for all leave-one-out folds (after selection)
%   tasks:     #contrasts string list
%   contrasts: #contrasts string list
%   subjects:  #subjects string list
%   labels:    #subjects vector (with values 1 or 2)
%   C:         structure as in "feature_name"
%
% Author: Jean Honorio
% E-mail: jhonorio@csail.mit.edu
function extract_features(ff_contrasts,root_output,tasks,contrasts,...
                          subjects,labels,fold_subject,scaling,C);

% Check input
warning off;
if length(tasks) ~= length(contrasts)
    error('Variables tasks and contrasts must be same length');
end

% Set method
if strcmp(C.method,'cluster')
    type = 'without_fold';
    num_features = 1;
elseif strcmp(C.method,'pca')
    type = 'without_fold_feature';
    num_features = C.components;
elseif strcmp(C.method,'ica')
    type = 'without_fold_feature';
    num_features = C.components;
elseif strcmp(C.method,'cubes')
    type = 'all';
    num_features = inf;
elseif strcmp(C.method,'discrim')
    type = 'without_fold';
    num_features = C.voxels;
elseif strcmp(C.method,'active')
    type = 'without_fold';
    num_features = C.voxels;
elseif strcmp(C.method,'slight')
    type = 'without_fold';
    num_features = C.voxels;
elseif strcmp(C.method,'rfesvm')
    type = 'without_fold';
    num_features = C.voxels;
else
    error('Not a valid feature selection method');
end

% Check fold array
provided = true;
if isempty(fold_subject)
    provided = false;
    fold_subject = 1:length(subjects);
end
check_fold_array(fold_subject);

% Load all subjects
data = load_contrasts(ff_contrasts,contrasts,subjects,scaling);
data = reshape(data,length(subjects),length(contrasts),[]);

% Extract features
for t = 1:length(tasks)
    root_features = fullfile(root_output,['features_',feature_name(C)],tasks{t});
    if strcmp(type,'all')
        
        % Load volume with labels (features)
        file_name = fullfile(root_features,[tasks{t},'_all.img']);
        fprintf('Loading %s...\n',file_name);
        hdr = spm_vol(file_name);
        vol = spm_read_vols(hdr);

        % Extract features
        folds(1,t).features = zeros(length(subjects),max(vol(:)));
        group = findvalues(vol(:));
        for f = 1:max(vol(:))
            idx = group(f).indices;
            for s2 = 1:length(subjects)
                folds(1,t).features(s2,f) = mean(data(s2,t,idx));
            end
        end
    elseif strcmp(type,'without_fold_feature')
        for s = 1:max(fold_subject)
            fprintf('Fold %d\n',s);
            idx_sf = find(fold_subject == s);
            fprintf('Subject');
            who = 'without';
            for sf = idx_sf
                fprintf(' %s',subjects{sf});
                who = [who '_' subjects{sf}];
            end
            fprintf('\n');
            idx_s = setdiff(1:length(subjects),idx_sf);
            
            % Load volume with labels (features)
            file_name = [tasks{t},'_',who];
            if length(file_name) > 150
                file_name = [file_name(1:145) '_more'];
            end

            % Extract features
            num_features_s = min(num_features,length(idx_s)-1);
            folds(s,t).features = zeros(length(subjects),num_features_s);
            for f = 1:num_features_s

                % Load each component (feature) as one volume
                file_name_f = fullfile(root_features,[file_name,'_',sprintf('%d',f),'.img']);
                fprintf('Loading %s...\n',file_name_f);
                hdr = spm_vol(file_name_f);
                vol = spm_read_vols(hdr);

                % Extract features
                idx = find(~isnan(vol(:)));
                total = sum(vol(idx));
                if abs(total) < 1e-8
                    error('Sum of weights too small');
                end
                vol(idx) = vol(idx) / total;
                for s2 = 1:length(subjects)
                    folds(s,t).features(s2,f) = dot(squeeze(data(s2,t,idx)),vol(idx));
                end
            end
        end
    elseif strcmp(type,'without_fold')        
        for s = 1:max(fold_subject)
            fprintf('Fold %d\n',s);
            idx_sf = find(fold_subject == s);
            fprintf('Subject');
            who = 'without';
            for sf = idx_sf
                fprintf(' %s',subjects{sf});
                who = [who '_' subjects{sf}];
            end
            fprintf('\n');

            % Load volume with labels (features)
            file_name = [tasks{t},'_',who];
            if length(file_name) > 150
                file_name = [file_name(1:145) '_more'];
            end
            file_name = fullfile(root_features,[file_name,'.img']);
            fprintf('Loading %s...\n',file_name);
            hdr = spm_vol(file_name);
            vol = spm_read_vols(hdr);
            
            % Extract features
            folds(s,t).features = zeros(length(subjects),num_features);
            group = findvalues(vol(:));
            for f = 1:max(vol(:))
                idx = group(f).indices;
                for s2 = 1:length(subjects)
                    folds(s,t).features(s2,f) = mean(data(s2,t,idx));
                end
            end
        end
    else
        error('This should not happen');
    end
end
file_name = fullfile(root_output,['features_',feature_name(C)],'extract.mat');
fprintf('Saving %s...\n',file_name);
if ~provided
    save(file_name,'folds','tasks','subjects','labels');
else
    save(file_name,'folds','tasks','subjects','labels','fold_subject');
end
fprintf('Done!\n');
