% [test_labels test_values C] = majvote(test_data,train_data,train_labels,tie);
%
% Run majority vote classifier on decision stumps
% Each decision stump is trained for minimizing training error
% Learns on train set, predicts test set
%   train_data, test_data: #instances by #features data matrices
%   train_labels:          #instances vector (with values 1 or 2)
%   tie:                   output class when ties (default 2)
%
% Author: Jean Honorio
% E-mail: jhonorio@csail.mit.edu
function [test_labels test_values C] = majvote(test_data,train_data,train_labels,tie);

% Check input
[num_instances num_features] = size(train_data);
if size(test_data,2) ~= num_features
    error('Training and testing set have different number of features');
end
if length(train_labels) ~= num_instances
    error('Number of labels is different from number of instances in training set');
end
if ~exist('tie','var')
    tie = 2;
end

% Learn on training data
zz = zeros(1,num_features);
C.theta = zz;
C.polarity = zz;
for f = 1:num_features
    min_epsilon = inf;
    min_theta = nan;
    min_polarity = nan;

    % Sort training data for feature across instances
    [sorted_data_f ind_data_f] = sort(train_data(:,f));

    % Fill temporary arrays
    tmp_class_1 = (train_labels(ind_data_f) == 1);
    tmp_class_2 = (train_labels(ind_data_f) == 2);

    % Compute error for every possible threshold
    sum_1_less = 0;
    sum_2_more = sum(tmp_class_2);
    sum_2_less = 0;
    sum_1_more = sum(tmp_class_1);
    for s = 1:num_instances - 1
        sum_1_less = sum_1_less + tmp_class_1(s);
        sum_2_more = sum_2_more - tmp_class_2(s);
        sum_2_less = sum_2_less + tmp_class_2(s);
        sum_1_more = sum_1_more - tmp_class_1(s);
        if sorted_data_f(s) ~= sorted_data_f(s + 1)
            theta = (sorted_data_f(s) + sorted_data_f(s + 1)) / 2;
            epsilon_neg = sum_1_less + sum_2_more;
            epsilon_pos = sum_2_less + sum_1_more;
            if epsilon_neg < epsilon_pos
                epsilon = epsilon_neg;
                polarity = -1;
            else
                epsilon = epsilon_pos;
                polarity = 1;
            end
            if epsilon < min_epsilon
                min_epsilon = epsilon;
                min_theta = theta;
                min_polarity = polarity;
            end
        end
    end
    C.theta(f) = min_theta;
    C.polarity(f) = min_polarity;
end

% Predict on test data
oo = ones(size(test_data,1),1);
test_values = sum((oo*C.polarity).*test_data < oo*(C.polarity.*C.theta),2) - num_features/2;
if tie == 2
    test_labels = 2 - double(test_values > 0);
else
    test_labels = 2 - double(test_values >= 0);
end
