% rank_voxels(ff_contrasts,root_output,task,contrast,...
%             subjects,labels,fold_subject,scaling,C);
%
% Rank all voxels for all leave-one-out folds
%   subjects:  #subjects string list
%   labels:    #subjects vector (with values 1 or 2)
%   C:         structure as follows:
% Decision stump:                C.method = 'stump'
% Gaussian classifier:           C.method = 'normal'
% Gaussian Naive Bayes on 3x3x3: C.method = 'gnb27'
% T-statistics for diff. means:  C.method = 'tstat'
%
% Author: Jean Honorio
% E-mail: jhonorio@csail.mit.edu
function rank_voxels(ff_contrasts,root_output,task,contrast,...
                     subjects,labels,fold_subject,scaling,C);

% Set method parameters
warning off;
if strcmp(C.method,'stump')
    type = 'voxel';
    classifier = @majvote;
    needs_discrete_labels = true;
elseif strcmp(C.method,'normal')
    type = 'voxel';
    classifier = @gnb;
    needs_discrete_labels = true;
elseif strcmp(C.method,'gnb27')
    type = 'neig27';
    classifier = @gnb;
    needs_discrete_labels = true;
elseif strcmp(C.method,'tstat')
    type = 'tstat';
    classifier = 0;
    needs_discrete_labels = true;
else
    error('Not a valid voxel ranking method');
end
has_continuous_labels = any(labels ~= 1 & labels ~= 2);
convert_labels = needs_discrete_labels & has_continuous_labels;
if convert_labels
    fprintf('\n*** RANKING WILL CONVERT CONTINUOUS TO DISCRETE LABELS ***\n\n');
end

% Check fold array
if isempty(fold_subject)
    fold_subject = 1:length(subjects);
end
check_fold_array(fold_subject);

% Create output directory if it doesn't exist
root_rankings = fullfile(root_output,['rankings_',C.method],task);
mkdir(root_rankings);

% Load all subjects
[data hdr] = load_contrasts(ff_contrasts,{contrast},subjects,scaling);
valid = ~isnan(squeeze(sum(data)));
hdr.fname = fullfile(root_rankings,'mask.img');
fprintf('Saving %s...\n',hdr.fname);
spm_write_vol(hdr,valid);

% Rank all voxels
[X Y Z] = size(valid);
for s = 0:max(fold_subject)
    if s == 0
        fprintf('All subjects\n');
        who = 'all';
        idx_s = 1:length(subjects);
    else
        fprintf('Fold %d\n',s);
        idx_sf = find(fold_subject == s);
        fprintf('Subject');
        who = 'without';
        for sf = idx_sf
            fprintf(' %s',subjects{sf});
            who = [who '_' subjects{sf}];
        end
        fprintf('\n');
        idx_s = setdiff(1:length(subjects),idx_sf);
    end
    file_name = [task,'_',who];
    if length(file_name) > 150
        file_name = [file_name(1:145) '_more'];
    end
    file_name = fullfile(root_rankings,[file_name,'.img']);
    if exist(file_name,'file')
        fprintf('...skipping, computed before\n');
        continue;
    end
    fprintf('xxxxxxxxxx');
    min_sigma = nan(X,Y,Z);
    label_train = labels(idx_s);
    % Convert continuous label to a discrete label (low/high)
    if convert_labels
        cut = mean(label_train); % median did not give good results
        label_train = (label_train >= cut) + 1;
    end
    if strcmp(type,'voxel')

        % Classification on single voxel value
        for x = 1:X
            for y = 1:Y
                fprintf('\b\b\b\b\b\b\b\b\b\bx=%02d, y=%02d',x,y);
                for z = 1:Z
                    if valid(x,y,z)
                        X_train = data(idx_s,1,x,y,z);
                        label_pred = classifier(X_train,X_train,label_train);
                        min_sigma(x,y,z) = sum(label_pred(:) == label_train(:));
                    end
                end
            end
        end
        min_sigma = min_sigma / length(idx_s);
    elseif strcmp(type,'neig27')
        
        % Classification on the 3x3x3 voxel neighborhood
        for x = 2:X-1
            for y = 2:Y-1
                fprintf('\b\b\b\b\b\b\b\b\b\bx=%02d, y=%02d',x,y);
                for z = 2:Z-1
                    if valid(x,y,z)
                        X_train = reshape(data(idx_s,1,x-1:x+1,y-1:y+1,z-1:z+1),length(idx_s),[]);
                        X_train = X_train(:,reshape(valid(x-1:x+1,y-1:y+1,z-1:z+1),1,[]));
                        label_pred = classifier(X_train,X_train,label_train);
                        min_sigma(x,y,z) = sum(label_pred(:) == label_train(:));
                    end
                end
            end
        end
        min_sigma = min_sigma / length(idx_s);
    elseif strcmp(type,'tstat')
        
        % T-statistics
        label_train1 = find(label_train == 1);
        label_train2 = find(label_train == 2);
        n1 = length(label_train1);
        n2 = length(label_train2);
        for x = 1:X
            for y = 1:Y
                fprintf('\b\b\b\b\b\b\b\b\b\bx=%02d, y=%02d',x,y);
                for z = 1:Z
                    if valid(x,y,z)
                        X_train = data(idx_s,1,x,y,z);
                        mu1 = mean(X_train(label_train1));
                        mu2 = mean(X_train(label_train2));
                        var1 = var(X_train(label_train1),1);
                        var2 = var(X_train(label_train2),1);
                        min_sigma(x,y,z) = abs(mu1 - mu2) / sqrt(var1 / n1 + var2 / n2);
                    end
                end
            end
        end
    else
        error('This should not happen');
    end
    fprintf('\n');

    % Save output
    hdr.fname = file_name;
    fprintf('Saving %s...\n',hdr.fname);
    spm_write_vol(hdr,min_sigma);
end
fprintf('Done!\n');
